#! /usr/bin/env python
#
# This MMTK/Python script is used to generate a pair of "Lego"-block like
# molecules. It is a somewhat improved version of what I described in my
# paper for MNT6.
# written by Peter McCluskey (pcm@rahul.net)
#

from MMTK import *
from MMTK.Database import BlueprintHook
from MMTK.Hooks import Hook
from MMTK.Geometry import *
from MMTK.Joiner import Joiner
from MMTK.Trajectory import StandardLogOutput
from MMTK.ForceFields import GenericForceField
from MMTK import Units
from MMTK import Positioner
from MMTK.CompositeObject import ConformationError
import sys

ff = GenericForceField('mm3')
world = InfiniteUniverse(ff)
block1 = Group('diamondoid_fragment')
join = Joiner(block1)
width = 5
join.repeatGroup('hookeast', 'hookwest', width)
join.repeatGroup('hooksouth', 'hooknorth', 6, 1, [180*Units.deg]*width)
join.repeatGroup('hookup', 'hookdown', 2)

result = join.getResult()
result.setPositions(ff)
#print '2.bonds_to_0',len(result.bonds_to_atom[id(result.atoms[0])])

for c_center in ('diamond4carbon.diamond4carbon37.C1','diamond4carbon.diamond4carbon52.C3'):
    a_hole = result.findAtomByName(c_center)
    surface_point = Positioner.findNearestPointOnSurface(result, a_hole)
    surface_vector = surface_point - a_hole.position()
    dist_to_surface = surface_vector.length()
    surface_dir = surface_vector.normal()
    cylinder1 = Cylinder(a_hole.position() - 0.1*Units.Ang*surface_dir,\
		     2.5*Units.Ang, (1*Units.Ang+dist_to_surface)*surface_dir)

    atoms_in_cyl1 = result.atomsWithinRegion(cylinder1)
    result.removeAtoms(atoms_in_cyl1, None)

c_below = [result.findAtomByName('diamond4carbon.diamond4carbon7.C2'),\
	   result.findAtomByName('diamond4carbon.diamond4carbon22.C4')]
nitrogens = [result.generateNewAtom('N'), result.generateNewAtom('N')]
result.replaceAtoms(c_below, nitrogens)
result.fillDanglingBonds(ff, 'N', 8, 3)

def filter2n(atom):
    if atom.symbol != 'C':
	return 0
    bonded_to = atom.bondedTo()
    countn = 0
    for a in bonded_to:
	if a.symbol == 'N':
	    countn = countn + 1
    return countn == 2

carbons_bonded_to_2_ns = filter(filter2n , result.atomList())
selected_ns = []
replace_cs = []
for a1 in carbons_bonded_to_2_ns:
    for a2 in a1.bondedTo():
	if a2.symbol == 'C':
	    c = a2
    a1_pos = a1.position()
    for a2 in a1.bondedTo():
	if a2.symbol == 'N':
	    x_product = (a1_pos - c.position()).cross(a2.position() - a1_pos)
	    if surface_dir * x_product > 0:
		selected_ns.append(a2)
    replace_cs.append(result.generateNewAtom('C'))
result.replaceAtoms(selected_ns, replace_cs, ff)
for i in range(len(carbons_bonded_to_2_ns)):        # make double bonds:
    result.attachBondBetween(carbons_bonded_to_2_ns[i], replace_cs[i])

hook3_list = []
for db in result.danglingbonds:
    for n in nitrogens:
	if n in db.a1.bondedTo():
	    hook3_list.append(db)
print 'hook3_list',hook3_list
join2 = Joiner(result)
cn = Group('cyanide')
for db in hook3_list:
    join2.joinGroups(Hook((result, [db])), cn.hook1, 0)
    result = join2.getIntermediateResult()
result = join2.getResult()
result.setPositions(ff)

result.removeAtoms([
    'diamond4carbon.diamond4carbon0.diamond4carbon58.C4',
    'diamond4carbon.diamond4carbon0.diamond4carbon59.C3',
    'diamond4carbon.diamond4carbon0.diamond4carbon59.C2',
    'diamond4carbon.diamond4carbon0.diamond4carbon28.C4',
    'diamond4carbon.diamond4carbon0.diamond4carbon29.C3',
    'diamond4carbon.diamond4carbon0.diamond4carbon29.C2',
    'diamond4carbon.diamond4carbon0.diamond4carbon59.C4',
    'diamond4carbon.diamond4carbon0.diamond4carbon29.C4',
    'diamond4carbon.diamond4carbon0.diamond4carbon54.C4',
    'diamond4carbon.diamond4carbon0.diamond4carbon59.C1',
    'diamond4carbon.diamond4carbon0.diamond4carbon29.C2',
    'diamond4carbon.diamond4carbon0.diamond4carbon29.C3',
    'diamond4carbon.diamond4carbon0.diamond4carbon28.C4',
    'diamond4carbon.diamond4carbon0.diamond4carbon57.C4',
    'diamond4carbon.diamond4carbon0.diamond4carbon57.C4',
    'diamond4carbon.diamond4carbon0.diamond4carbon58.C2',
    'diamond4carbon.diamond4carbon0.diamond4carbon58.C3',
    'diamond4carbon.diamond4carbon0.diamond4carbon0.C1',
    'diamond4carbon.diamond4carbon0.diamond4carbon0.C3',
    'diamond4carbon.diamond4carbon0.diamond4carbon1.C1',
    'diamond4carbon.diamond4carbon0.diamond4carbon30.C1',
    'diamond4carbon.diamond4carbon0.diamond4carbon5.C1',
    'diamond4carbon.diamond4carbon0.diamond4carbon30.C3',
    'diamond4carbon.diamond4carbon0.diamond4carbon30.C2',
    'diamond4carbon.diamond4carbon0.diamond4carbon31.C1',
    'diamond4carbon.diamond4carbon0.diamond4carbon2.C1',
    'diamond4carbon.diamond4carbon0.diamond4carbon0.C2',
    'diamond4carbon.diamond4carbon0.diamond4carbon1.C2',
    'diamond4carbon.diamond4carbon0.diamond4carbon1.C3',
    'diamond4carbon.diamond4carbon0.diamond4carbon0.C4'
    ], None)
try:
    result.hydrogenateDanglingBonds(ff)
except ConformationError,msg:
    from MMTK import RasMol
    ras_session = RasMol.RasMol(result, 1)
    ras_session.RefreshScreen()
    ras_session.Interact()
    raise ConformationError, msg
    

result2 = result.clone()	# shouldn't this be Molecule? (no clone yet)
result2.name = 'd42'
face2 = []
for a in result2.atomList():
    if a.symbol == 'N':
	bonded = a.bondedTo()
	if len(bonded) == 1 and bonded[0].symbol == 'C':
	    face2.append(a)

face2 = [face2[1], face2[0], face2[2]]
#print 'face2',face2
coordlist = Positioner.PositionAdjacent(result, carbons_bonded_to_2_ns[0:3],\
					result2, face2, [2*Units.Ang]*3)

from MMTK.ConfigIO import OutputFile
fmt = OutputFile('resultl5.1.pdb')
fmt.write(result)
fmt = OutputFile('resultl5.2.pdb')
fmt.write(result2)
fmt.close()

result.knob1 = [result.findAtomByName('diamond4carbon.diamond4carbon3.N'),\
		result.findAtomByName('diamond4carbon.diamond4carbon1.N'),\
		result.findAtomByName('diamond4carbon.diamond4carbon2.N')]

result.knob2 = [result.findAtomByName('diamond4carbon.diamond4carbon5.N'),\
		result.findAtomByName('diamond4carbon.diamond4carbon4.N'),\
		result.findAtomByName('diamond4carbon.diamond4carbon6.N')]

result.hole1 = [result.findAtomByName('diamond4carbon.diamond4carbon0.diamond4carbon53.C2'),\
		result.findAtomByName('diamond4carbon.diamond4carbon0.diamond4carbon51.C2'),\
		result.findAtomByName('diamond4carbon.diamond4carbon0.diamond4carbon56.C2')]

result.hole2 = [result.findAtomByName('diamond4carbon.diamond4carbon0.diamond4carbon33.C4'),\
		result.findAtomByName('diamond4carbon.diamond4carbon0.diamond4carbon31.C4'),\
		result.findAtomByName('diamond4carbon.diamond4carbon0.diamond4carbon36.C4')]

save(result,'result.lego7')

world.addObject(result)
world.addObject(result2)

fmt = OutputFile('resultlw.pdb')
fmt.write(world)
fmt.close()

if 0:
    world.view()
else:
    from MMTK import RasMol
    ras_session = RasMol.RasMol(world, 1)
    for coords in coordlist:
	ras_session.AddPolyLines(coords)
    ras_session.RefreshScreen()
    ras_session.Interact()
