# This module defines environment objects for universes.
#
# Written by Konrad Hinsen
# last revision: 1998-10-27
#

import Numeric

#
# The environment object base class
#
class EnvironmentObject:

    is_environment_object = 1

    def checkCompatibilityWith(self, other):
        pass

# Type check

def isEnvironmentObject(object):
    return hasattr(object, 'is_environment_object')

#
# Nose thermostat class
#
class NoseThermostat(EnvironmentObject):

    def __init__(self, temperature, relaxation_time = 0.2):
        self.parameters = Numeric.array([temperature, relaxation_time])
        self.coordinates = Numeric.array([0., 0.])

    def setTemperature(self, temperature):
        self.parameters[0] = temperature

    def setRelaxationTime(self, t):
        self.parameters[1] = t

    def checkCompatibilityWith(self, other):
        if other.__class__ is NoseThermostat:
            raise ValueError, "the universe already has a thermostat"

#
# Andersen barostat class
#
class AndersenBarostat(EnvironmentObject):

    def __init__(self, pressure, relaxation_time = 1.5):
        self.parameters = Numeric.array([pressure, relaxation_time])
        self.coordinates = Numeric.array([0.])

    def setPressure(self, pressure):
        self.parameters[0] = pressure

    def setRelaxationTime(self, t):
        self.parameters[1] = t

    def checkCompatibilityWith(self, other):
        if other.__class__ is AndersenBarostat:
            raise ValueError, "the universe already has a barostat"

