# A Lennard-Jones force fields for simple liquids
#
# Written by Konrad Hinsen
# last revision: 1999-2-23
#

from NonBondedInteractions import LJForceField
from Scientific.Geometry import Vector
import copy

#
# The force field class
#
class LennardJonesForceField(LJForceField):

    def __init__(self, cutoff = None):
	self.arguments = (cutoff, )
	LJForceField.__init__(self, 'LJ', cutoff)
        self.lj_14_factor = 1.

    def ready(self):
        return 1

    def collectParameters(self, universe, global_data):
	if not hasattr(global_data, 'lj_parameters'):
	    parameters = {}
	    for o in universe:
		for a in o.atomList():
		    parameters[a.symbol] = (a.LJ_energy, a.LJ_radius)
	    global_data.lj_parameters = parameters

    def _atomType(self, object, atom, global_data):
        return atom.symbol

    def _ljParameters(self, type, global_data):
	return global_data.lj_parameters[type]

    def evaluator(self, universe, subset1, subset2, global_data):
        self.collectParameters(universe, global_data)
	return LJForceField.evaluator(self, universe, subset1, subset2,
				      global_data)
