# This module implements a class that connects Groups together
#
# Written by Peter McCluskey (pcm@rahul.net)
#

from Hooks import Hook, isHook, HookList, isHookList, mergeHooklist
import Database, Bonds, ChemicalObjects
import DictWithDefault, types
import Units
import Utility
import pprint
import os
from time import clock

times0 = 0
times1 = 0
times2 = 0
times3 = 0
times4 = 0
times5 = 0
times6 = 0
times7 = 0
times8 = 0

global_subgroupmap = {}

class Joiner:
  def __init__(self, group):
      self._result = None
      self._group = group
      #if hasattr(group, 'subgroups'):
	#  print 'Joiner.__init__'
	#  _printsublist(group)
      self._group_name = group.name
      self._count = 0
      self._sublist = []
      self._is_in_sublist = DictWithDefault.DictWithDefault(0)
      self._newSubgroupMap = global_subgroupmap
      self._collapsable = DictWithDefault.DictWithDefault(0)

  def _mark(self, group, cloned):
      if hasattr(group, 'joiner') and group.joiner is not None:
	  if group.joiner is self:
	      return
	  if not cloned:
	      msg = 'group "' + group.name + '" passed to 2 Joiners'
	      raise RuntimeError,msg
	  unPropagateDataUp(group)
      group.joiner = self

  def getResult(self, shorten_names = 1):
      if self._sublist and shorten_names:
	  self.ShortenNames()
      r = self._result
      propagateDataUp(r)
      #print 'getResult:'
      #_printsublist(r)
      self._result = None
      return r

  def getIntermediateResult(self):
      return self._result

			# joining_bond_dihedrals is list of angles (radians);
			# it's only usefull for single-bond hooks
  def joinGroups(self, hook1, hook2, need_to_rename_group1_atoms = 1,
		 joining_bond_dihedrals = None, min_copy_atom = 0):
    #print 'Join:',hex(id(hook1)),hex(id(hook2)),hook1, hook2
    newSubgroupMap = self._newSubgroupMap
    nbonds = len(hook1)
    if nbonds != len(hook2):
	raise ValueError, 'Hooks have different numbers of bonds (' +\
	      `nbonds` + ' vs. ' + `len(hook2)` + ')'
    if nbonds < 1:
	raise ValueError, 'Hooks have no bonds to attach'

    global times1
    global times3
    global times4
    global times5
    global times6
    global times7
    global times8
    global times0
    global times2

    times0 = times0 - clock()
    bonds1 = hook1.getBonds()[:]
    bonds2 = hook2.getBonds()[:]
    new_bonds = []
    new_dihedrals = []
    group1 = hook1.getGroup()
    self._mark(group1, need_to_rename_group1_atoms)
    group2 = hook2.getGroup()
    self._mark(group2, 1)
    print 'join2',hex(id(group1)),hex(id(group2))
    name2 = group2.name
    if len(group1.danglingbonds + group2.danglingbonds) - (len(bonds1 + bonds2)) == 0:
	result = ChemicalObjects.Molecule(None)
    else:
	result = ChemicalObjects.Group(None)
    print 'create new obj',hex(id(result))
    result.parent = None
    times1 = times1 - clock()
    times1 = times1 + clock()
    if hasattr(group1, 'dihedrals'):
	group1_dihedrals = group1.dihedrals[:]
    else:
	group1_dihedrals = []
    if hasattr(group2, 'dihedrals'):
	group2_dihedrals = group2.dihedrals[:]
    else:
	group2_dihedrals = []

    times3 = times3 - clock()
    print 'Clone1'
    if need_to_rename_group1_atoms:
	subgroup1 = self._cloneSubgroups(group1, result, 1)
	#self._cloneOriginalSubgroups(subgroup1, group1)
	self._updateAttachments(subgroup1)
	#_printAttachments(subgroup1)
    else:
	subgroup1 = group1
	if not hasattr(subgroup1, 'subgroups'):
	    subgroup1.subgroups = []
	if not hasattr(subgroup1, 'bonds'):
	    subgroup1.bonds = []
	subgroup1.parent = result
	print hex(id(subgroup1)),'.parent = ',hex(id(result)),',',hex(id(result.parent))
    print 'Clone2',hex(id(group1)),hex(id(subgroup1))
    subgroup2 = self._cloneSubgroups(group2, result, 0)
    #self._cloneOriginalSubgroups(subgroup2, group2)
    self._updateAttachments(subgroup2)
    #_printAttachments(subgroup2)
    times3 = times3 + clock()
    print 'Clone2done'
    result.subgroups = [subgroup1, subgroup2]
    #result.original_subgroups = [subgroup1, subgroup2]
    if subgroup1.name == subgroup2.name:
	result.name = subgroup1.name + '_b'
	subgroup1.name = subgroup1.name + '0'
	subgroup2.name = subgroup2.name + '1'
    else:
	result.name = group1.name + '_' + group2.name
    #_printsublist(subgroup1)
    #print
    #_printsublist(subgroup2)
    print 'result.name',hex(id(result)),result.name,hex(id(subgroup1)),hex(id(subgroup2)),subgroup1.name[-32:],subgroup2.name[-32:], need_to_rename_group1_atoms

    new_hook1 = DictWithDefault.DictWithDefault(None)
    new_hook2 = DictWithDefault.DictWithDefault(None)
    times2 = times2 - clock()
    #print 'Used5:',os.popen("ps -ux | grep python").read()
    group1_hooks = filterHooks(group1.hooks, result, subgroup1, hook1,\
		       new_hook1, newSubgroupMap, need_to_rename_group1_atoms)
    group2_hooks = filterHooks(group2.hooks, result, subgroup2, hook2, \
			       new_hook2, newSubgroupMap)
    #print 'Used6:',os.popen("ps -ux | grep python").read()
    times2 = times2 + clock()
    result.hooks = HookList(group1_hooks + group2_hooks)

    for attr in group2.__dict__.keys():
	h = new_hook2[id(getattr(group2, attr))]
	if h is not None:
	    setattr(result, attr, h)
    for attr in group1.__dict__.keys():
	h = new_hook1[id(getattr(group1, attr))]
	if h is not None:
	    setattr(result, attr, h)	# ?? handle dups ??

    skip_attrs = ('atoms', 'bonds', 'dihedrals', 'danglingBonds',\
		  'amber_atom_type', 'mm3_atom_type', 'joiner', \
		  'hooks', 'name', 'subgroups', \
		  'atom_references', 'parent', 'bonds_to_atom')

    times0 = times0 + clock()

    attrlist = vars(group1).keys()
    for attr in attrlist:
	if attr in skip_attrs:
	    continue
	if hasattr(group2, attr):
	    attr1 = getattr(group1, attr)
	    attr2 = getattr(group2, attr)
	    try:
		attr_sum = attr1 + attr2
	    except:
		if attr1 is None:
		    attr_sum = attr2
		elif attr2 is None:
		    attr_sum = attr1
		elif type(attr1) is type({}):
		    attr_sum = {}
		    for key in attr1.keys():
			attr_sum[key] = attr1[key]
		    for key in attr2.keys():
			try:
			    old = attr_sum[key]
			    if old != attr2[key]:
				msg = `attr`+'['+`key`+'] has 2 values:"'+`old`+'" and "' + `attr2[key]`+'"'
				raise ValueError,msg
			except:
			    attr_sum[key] = attr2[key]
		elif type(attr1) != types.InstanceType:
			print 'error - cannot combine types:',attr1,attr2
			continue
		elif attr2.__class__ is Hook and attr1.__class__ is HookList:
			attr_sum = HookList([])
			for h in attr1:
			    new_h = new_hook1[id(h)]
			    if new_h is None:
				print '[]2 no new_hook for',hex(id(h)),h,attr2,attr1
				new_h = h
			    else:
				attr_sum.append(new_h)
			attr_sum.append(new_hook2[id(attr2)])
			pos = result.hooks.index(new_hook2[id(attr2)])
			result.hooks[pos] = attr_sum
		elif attr1.__class__ is Bonds.DanglingBond:
		    continue
		elif attr1.__class__ is Bonds.Dihedral:
		    continue
		elif attr1.__class__ is Bonds.Bond:
		    continue
		elif attr1.__class__ is Hook:
		    attr_sum = HookList([new_hook1[id(attr1)], new_hook2[id(attr2)]])
		    if attr_sum[0] is None:
			if attr_sum[1] is None:
			    continue
			attr_sum = attr_sum[1]
			pos = result.hooks.index(attr_sum)
			result.hooks[pos] = attr_sum
		    elif attr_sum[1] is None:
			attr_sum = attr_sum[0]
			pos = result.hooks.index(attr_sum)
			result.hooks[pos] = attr_sum
		    else:
			pos = result.hooks.index(new_hook1[id(attr1)])
			result.hooks[pos] = attr_sum
			result.hooks.remove(new_hook2[id(attr2)])
		elif attr1.__class__ is Database.GroupType:
		    attr_sum = attr1	# should construct MoleculeType sometime
		elif attr1.__class__ is ChemicalObjects.Atom:
		    continue
		    #attr_sum = attr1	# group2.atoms will be handled later
		else:
		    print 'error combining ',attr,type(attr1),type(attr2),attr1.__class__,attr2.__class__
		    attr_sum = attr1 + attr2
		    continue
	    if isHookList(attr1) and isHookList(attr2) and len(attr_sum) > 0:
		attr1_sum = HookList([])
		for h in attr1:
		    new_h = new_hook1[id(h)]
		    if new_h is None:
			print '[]1 no new_hook for',hex(id(h)),h,attr1
			new_h = h
		    attr1_sum.append(new_h)
		attr2_sum = HookList([])
		for h in attr2:
		    new_h = new_hook2[id(h)]
		    if new_h is None:
			print '2a no new_hook for',hex(id(h)),h,attr2
			new_h = h
		    attr2_sum.append(new_h)
		if hook1.matchList(mergeHooklist(attr1, new_hook1)):
		    attr_sum = attr2_sum
		elif hook2.matchList(mergeHooklist(attr2, new_hook2)):
		    attr_sum = attr1_sum
		else:
		    attr_sum = HookList(attr1_sum + attr2_sum)
		#print attr,'combine2',len(attr_sum),len(attr1),len(attr2) #,attr_sum
	    setattr(result, attr, attr_sum)
	else:
	    attr1 = getattr(group1, attr)
	    #if attr == 'hookup_extra': print 'hookup_extra attr1',getattr(group1, attr)
	    if isHookList(attr1) and len(attr1) > 0:
		attr1_sum = HookList([])
		for h in attr1:
		    new_h = new_hook1[id(h)]
		    if new_h is None:
			print 'b no new_hook for',hex(id(h)),h,attr1
			new_h = h
		    attr1_sum.append(new_h)
		#print 'handle 1 hook',attr,attr1,attr1_sum
		if attr1_sum[0] is None:
		    continue
		if hook1.matchList(mergeHooklist(attr1, new_hook1)):
		    continue
		else:
		    attr1 = attr1_sum
	    setattr(result, attr, attr1)
    for attr in group2.__dict__.keys():
	#if attr == 'hookup_extra': print 'hookup_extra attr2',getattr(group2, attr)
	if hasattr(group1, attr): continue
	if attr in skip_attrs: continue
	attr2 = getattr(group2, attr)
	if isHookList(attr2) and len(attr2) > 0:
	    attr2_sum = HookList([])
	    for h in attr2:
		new_h = new_hook2[id(h)]
		if new_h is None:
		    print 'c no new_hook for',hex(id(h)),h,attr2
		    #new_h = h
		    attr2_sum.append(new_h) # ??
	    #print 'handle 2 hook',attr,attr2,attr2_sum
	    if attr2_sum[0] is None:
		continue
	    if hook2.matchList(mergeHooklist(attr2, new_hook2)):
		continue
	    else:
		attr2 = attr2_sum
	setattr(result, attr, attr2)

		# remove dangling bonds used to join
    group1_danglingbonds = Bonds.BondList(group1.danglingbonds)
    group2_danglingbonds = Bonds.BondList(group2.danglingbonds)
    for i in range(nbonds):
	#print 'remove',hex(id(bonds1[i])),hex(id(bonds2[i])),bonds1[i].a1.name,bonds2[i].a1.name
	group1_danglingbonds.remove(bonds1[i])
	group2_danglingbonds.remove(bonds2[i])

    old_atom1 = {}
    old_atom2 = {}
    new_bonds1 = DictWithDefault.DictWithDefault(None)
    new_bonds2 = DictWithDefault.DictWithDefault(None)

    times4 = times4 - clock()

    new_atoms1 = DictWithDefault.DictWithDefault(None)
    new_atoms2 = DictWithDefault.DictWithDefault(None)

    if need_to_rename_group1_atoms:
	result.atoms = []
	for a in group1.atomList():
	    new_name = a.name # check for clashes?
	    new_atom = a.cloneAndRename(result, new_name)
	    setattr(result, new_name, new_atom)
	    new_atom.parent = result
	    if hasattr(a, 'parent_subgroup'):
		new_atom.parent_subgroup = _find_equiv_subgroup(subgroup1, a.parent_subgroup, newSubgroupMap, 1)
	    else:
		new_atom.parent_subgroup = subgroup1
	    new_atoms1[id(a)] = new_atom
	    #print '1new_atom',hex(id(new_atom)),'from',hex(id(a))
	    old_atom1[id(new_atom)] = a
	    result.addAtom(new_atom)

	replaceAtom(subgroup1, new_atoms1, new_bonds1)
	replaceAtom(bonds1, new_atoms1, new_bonds1)
	replaceAtom(group1_danglingbonds, new_atoms1, new_bonds1)
	replaceAtom(group1_dihedrals, new_atoms1, None)
    else:
	result.atoms = group1.atoms[:] # want to just assign here for speed, but problems w using group1.atoms later
	for a in group1.atoms:
	    a.parent = result
	    setattr(result, a.name, a)
	    if hasattr(a, 'parent_subgroup'):
		a.parent_subgroup = _find_equiv_subgroup(subgroup1, a.parent_subgroup, newSubgroupMap, 0)
	    else:
		a.parent_subgroup = subgroup1
    times4 = times4 + clock()

    for a in group2.atomList():
	new_name = a.name # check for clashes?
	new_atom = a.cloneAndRename(result, new_name)
	if hasattr(a, 'parent_subgroup'):
	    new_atom.parent_subgroup = _find_equiv_subgroup(subgroup2, a.parent_subgroup, newSubgroupMap, 0)
	else:
	    new_atom.parent_subgroup = subgroup2
	#print 'new_atom2',hex(id(new_atom)),a.name,'parent_subgroup',hex(id(subgroup2))
	new_atoms2[id(a)] = new_atom
	#print '2new_atom',hex(id(new_atom)),'from',hex(id(a))
	old_atom2[id(new_atom)] = a
	result.addAtom(new_atom)
	setattr(result, new_name, new_atom)# ? overwrites when group1 is group2
	new_atom.parent = result
    #print 'Used7:',os.popen("ps -ux | grep python").read()
    times6 = times6 - clock()
    print 'start new_bonds2'
    times6 = times6 + clock()
    times5 = times5 - clock()
    replaceAtom(subgroup2, new_atoms2, new_bonds2)
    replaceAtom(bonds2, new_atoms2, new_bonds2)
    times5 = times5 + clock()
    replaceAtom(group2_danglingbonds, new_atoms2, new_bonds2)
    replaceAtom(group2_dihedrals, new_atoms2, None)

    for i in range(nbonds):
	dbond1 = bonds1[i]
	dbond2 = bonds2[i]
	new_bonds.append(Bonds.Bond((dbond1.a1, dbond2.a1)))
	#print 'join new bond',dbond1, dbond2, hex(id(dbond1.a1)),hex(id(dbond2.a1)), hex(id(new_bonds[-1]))
	_addDihedral(new_dihedrals, dbond1, dbond2.a1)
	_addDihedral(new_dihedrals, dbond2, dbond1.a1)
	try:
	    angle = joining_bond_dihedrals[i]
	    jb1 = dbond1.constraint.b2
	    jb2 = dbond2.constraint.b2
	except (IndexError, AttributeError, TypeError): pass
	else:
	    join_d = Bonds.Dihedral((jb1, jb2, new_bonds[-1], angle))
	    print 'joining_bond_dihedral',hex(id(join_d))
	    new_dihedrals.append(join_d)
    result.danglingbonds = Bonds.BondList(group1_danglingbonds + group2_danglingbonds)

    result.bonds_to_atom = {}
    if need_to_rename_group1_atoms:
	replaceBond(subgroup1, new_bonds1)
	print 'replaceBond subgroup1 done'
	replaceBond(group1_dihedrals, new_bonds1)
	replaceHookBonds(group1_hooks, new_bonds1)
	replaceBondAngleLists(group1.atomList(), new_atoms1)
	if hasattr(group1, 'bonds_to_atom'):
	    for (key, blist) in group1.bonds_to_atom.items():
		blist = blist[:]
		replaceBond(blist, new_bonds1)
		result.bonds_to_atom[id(new_atoms1[key])] = blist
    elif hasattr(group1, 'bonds_to_atom'):
	for (key, blist) in group1.bonds_to_atom.items():
	    result.bonds_to_atom[key] = blist
	
    replaceBondAngleLists(group2.atomList(), new_atoms2)
    replaceBond(group2_dihedrals, new_bonds2)
    replaceBond(subgroup2, new_bonds2)
    replaceHookBonds(group2_hooks, new_bonds2)
    if hasattr(group2, 'bonds_to_atom'):
	for (key, blist) in group2.bonds_to_atom.items():
	    blist = blist[:]
	    replaceBond(blist, new_bonds2)
	    result.bonds_to_atom[id(new_atoms2[key])] = blist

    #print 'Used8:',os.popen("ps -ux | grep python").read()
    attrlist = ('amber_atom_type','amber_charge','mm3_atom_type','mm3_charge')
    attr_ref = []
    attr_ref1 = []
    attr_ref2 = []
    for attr in attrlist:
	setattr(result, attr, {})
	attr_ref.append(getattr(result, attr))
	try:
	    attr_ref1.append(getattr(group1, attr))
	except AttributeError: attr_ref1.append(None)
	try:
	    attr_ref2.append(getattr(group2, attr))
	except AttributeError: attr_ref2.append(None)
    atomlist = result.atomList()
    len1 = len(group1.atoms)
    start_group2_index = len1
    if need_to_rename_group1_atoms:
	start_index = 0
    else:
	start_index = len1
	for i in range(len1):
	    a = atomlist[i]
	    if a.index is not None and a.index >= start_group2_index:
		start_group2_index = a.index + 1 # hack, used if atom removed
    result.atom_references = {}
    for i in range(0, min_copy_atom):
	for attr in attrlist:
	    result.setAtomProperty(atomlist[i], attr,\
			   group1.getAtomProperty(group1.atoms[i], attr))
    for i in range(min_copy_atom, len(atomlist)):
	#try:
	#    print 'index',i,'of',len(atomlist),len1,atomlist[i].index
	#except AttributeError: print 'n/a'
	a = atomlist[i]
	joinedkey = result.getReferenceFromIndex(a, i)
	result.setIndex(a, i)
	for j in range(len(attrlist)):
	    attr = attrlist[j]
	    ref = attr_ref[j]
	    if i < len1:
		attr1 = attr_ref1[j]
		if attr1 is None:
		    continue
		try:
		    search_atom1 = old_atom1[id(a)]
		except KeyError: search_atom1 = a
		ref1 = group1.getReferenceFromIndex(search_atom1, i)
		try:
		    val = attr1[ref1]
		except KeyError:
		    continue
	    else:
		attr2 = attr_ref2[j]
		if attr2 is None:
		    continue
		try:
		    search_atom2 = old_atom2[id(a)]
		except KeyError: search_atom2 = a
		ref2 = group2.getReferenceFromIndex(search_atom2, i - len1)
		try:
		    val = attr2[ref2]
		except KeyError:
		    continue
	    ref[joinedkey] = val
	    if 0:
		print hex(id(a)),a.fullName(),'attr',attr,'set to',val,hex(id(result)),hex(id(ref)),joinedkey

    result.bonds = Bonds.BondList(new_bonds)
    result.dihedrals = group1_dihedrals +\
		       Bonds.DihedralAngleList(new_dihedrals) +\
		       group2_dihedrals
    print 'Hook times',times0,times1,times2,times3,times4,times5,times6
    #if not need_to_rename_group1_atoms:
    #	delSubgroups(group1)
    if 0:
	for a in result.atoms[0:4]:
	    _checkSublistConsistency(a.parent_subgroup)
    #(group1 is self._result)
    self._result = result
    self._group = result
    self._collapsable[id(result)] = 1
    subgroup2.attach_subgroups = [subgroup1]
    self.HandleNames(result.subgroups)
    #_printSubs(result.subgroups)
    #print 'result.bonds'
    #for b in result.bonds:
	#print hex(id(b)),b
    #print 'Used9:',os.popen("ps -ux | grep python").read()

  def repeatGroup(self, hook1name, hook2name, n, need_to_rename_atoms = 1,\
		  joining_bond_dihedrals = None):
    result = self._group
    group_name = self._group.name
    hook2 = getattr(self._group, hook2name)
    if type(hook2) is type([]) or isHookList(hook2):
	hook2 = Hook((self._group, mergeHooklist(hook2)))
    for i in range(1, n):
	hook1 = getattr(result, hook1name)
	if type(hook1) is type([]) or isHookList(hook1):
	    print 'merging hook'
	    hook1 = Hook((result, mergeHooklist(hook1)))
	self.joinGroups(hook1, hook2, i == 1, joining_bond_dihedrals)
	result = self._result
	print 'repeatGroup',i, result.name, hook1name
    if need_to_rename_atoms:
	_checkSublistMap(result, self._newSubgroupMap)
    self._result = result

  def HandleNames(self, subgroups):
      for s in subgroups:
	  if self._is_in_sublist[id(s)]:
	      continue
	  elif not self._collapsable[id(s)]:
	      s.name = '%s%d' % (self._group_name, self._count)
	      self._count = self._count + 1
	      #print 'HandleNames',hex(id(s)),hex(id(self._result)),s.name
	      self._sublist.append(s)
	      self._is_in_sublist[id(s)] = 1
	  else:
	      #print 'HandleNames collapsable',hex(id(s)),hex(id(self._result)),s.name
	      self.HandleNames(s.subgroups)

  def ShortenNames(self):
    result = self._result
    result.name = self._group_name
    sublist = self._sublist
    #for i in range(len(sublist)):
	#print 'done w',i,sublist[i].name,hex(id(sublist[i])),hex(id(result))
    self._collapseSubGroupData(result, DictWithDefault.DictWithDefault(None))
    result.subgroups = sublist
    #_printSubs(result.subgroups)
    print 'resultlist',result.name,hex(id(result))
    #_printSubs(result.original_subgroups)
    #for a in result.atoms:
	#print hex(id(a)),a.fullName()
    #_printsublist(result)

  def _updateAttachments(self, group, depth = 1):
      for grp in group.subgroups:
	  self._updateAttachments(grp, depth + 1)
      if hasattr(group, 'attach_subgroups'):
	  for i in range(len(group.attach_subgroups)):
	      grp = group.attach_subgroups[i]
	      while not _containsGroup(group, grp) \
		    and self._newSubgroupMap.has_key(id(grp)):
		  grp = self._newSubgroupMap[id(grp)]
		  #print ' '*depth,'_updateAttachment',fullSublistName(group.attach_subgroups[i])
		  #print ' '*depth,'to',fullSublistName(grp)
	      group.attach_subgroups[i] = grp

  def _collapseSubGroupData(self, grp, collapsed_map, depth = 1):
    for subgrp in grp.subgroups:
	if self._collapsable[id(subgrp)]:
	    self._collapseSubGroupData(subgrp, collapsed_map, depth + 1)
	    #print '  '*depth,'collapse',hex(id(subgrp)),subgrp.name,hex(id(self._result)),len(subgrp.bonds),len(self._result.bonds)
	    if subgrp.bonds:
		assert(not (subgrp.bonds[0] in self._result.bonds))
		self._result.bonds = self._result.bonds + subgrp.bonds
	    _reDoAttachedGroups(grp, subgrp, collapsed_map)
	    self._newSubgroupMap[id(subgrp)] = self._result
	    for s in subgrp.subgroups:
		s.parent = self._result
	    collapsed_map[id(subgrp)] = self._result

  def _cloneSubgroups(self, group, parent, leftside):
    subgroup = ChemicalObjects.Group(None)
    if self._newSubgroupMap.has_key(id(group)):
	#if newSubgroupMap.has_key((id(group),id(group))):
	#    print hex(id(newSubgroupMap[id(group)])),'overwrites',hex(id(newSubgroupMap[(id(group),id(group))]))
	self._newSubgroupMap[(id(group),id(group))] = self._newSubgroupMap[id(group)]
    try:
	subgroup.attach_subgroups = group.attach_subgroups[:]
    except AttributeError: pass
    self._newSubgroupMap[id(group)] = subgroup
    self._collapsable[id(subgroup)] = self._collapsable[id(group)]
    if leftside:
	self._is_in_sublist[id(subgroup)] = self._is_in_sublist[id(group)]
	if self._is_in_sublist[id(subgroup)]:
	    pos = self._sublist.index(group)
	    self._sublist[pos] = subgroup
    #print '_cloneSubgroups',hex(id(group)),hex(id(subgroup)),group.name,hex(id(self._newSubgroupMap))
    subgroup.atoms = group.atoms[:]
    try:
	bondlist = group.bonds
    except AttributeError:
	bondlist = []
    subgroup.bonds = Bonds.BondList(bondlist)
    subgroup.subgroups = []
    if hasattr(group,'subgroups'):
	for s in group.subgroups:
	    new_s = self._cloneSubgroups(s, subgroup, leftside)
	    subgroup.subgroups.append(new_s)
	    #print '_cloneSubgroups',hex(id(s)),hex(id(new_s)),fullSublistName(s)
	    #for a in s.atomList():
		#if hasattr(a,'parent_subgroup') and a.parent_subgroup is s:
		#    a.parent_subgroup = new_s
		#    print 'replce',hex(id(a)),a.name,hex(id(s)),hex(id(new_s))
    subgroup.name = group.name
    subgroup.parent = parent
    return subgroup

  def _cloneOriginalSubgroups(self, newgrp, oldgrp):
      if not hasattr(oldgrp, 'original_subgroups'):
	  newgrp.original_subgroups = []
	  return
      newgrp.original_subgroups = map(lambda s, m = self._newSubgroupMap : m[id(s)],\
				      oldgrp.original_subgroups)
      for i in range(len(oldgrp.original_subgroups)):
	  self._cloneOriginalSubgroups(newgrp.original_subgroups[i],\
				       oldgrp.original_subgroups[i])

def _printOrigSubgroups(obj, depth = 1):
    for o in obj.original_subgroups:
	print '  '*depth,len(o.bonds),len(o.joining_bonds),hex(id(o)),o.name
	_printOrigSubgroups(o, depth + 1)

def _checkSublistMap(obj, newSubgroupMap):
    if hasattr(obj, 'subgroups'):
	for s in obj.subgroups:
	    if not newSubgroupMap.has_key(id(s)):
		print 'warning - no map for',hex(id(s))
	    _checkSublistMap(s, newSubgroupMap)

def _printSubs(l, depth = 1):
    for s in l:
	if hasattr(s, 'bonds') and len(s.bonds) > 0:
	    print ' '*depth,len(s.bonds),hex(id(s)),s.name,hex(id(s.bonds[0]))#,fullSublistName(s)
	else: print ' '*depth,None,hex(id(s)),s.name,None#,fullSublistName(s)
	_printSubs(s.subgroups,depth+1)


def _updateOriginalGroupBonds(grp):
    print '_updateOriginalGroupBonds',hex(id(grp))
    for s in grp.original_subgroups:
	if hasattr(s, 'replacing_group') and not hasattr(s, 'joining_bonds'):
	    print '_updateOriginalGroupBond',hex(id(s))
	    t = s.replacing_bond_range
	    s.joining_bonds = s.replacing_group.joining_bonds[t[0]:t[1]]
	    s.bonds = s.joining_bonds
	    for b1 in s.bonds:
		print hex(id(b1)),b1.a1.fullName()[8:],b1.a2.fullName()[8:]
	_updateOriginalGroupBonds(s)

def propagateDataUp(obj):
    _printSubs(obj.subgroups)
    propagateBondDataUp(obj)
    #_updateOriginalGroupBonds(obj)
    propagateAtomDataUp1(obj.atoms)
    propagateAtomDataUp(obj)

def propagateBondDataUp(obj):
		# should verify that propagateDataUp not run twice on obj?
    try:
	bondlist = obj.bonds
    except AttributeError:
	bondlist = []
    #print 'propagateBondDataUp',fullSublistName(obj),len(bondlist),hasattr(obj,'joining_bonds')
    have_bond = DictWithDefault.DictWithDefault(0)
    for b in bondlist:
	have_bond[id(b)] = 1
	#print hex(id(b)),b.a1.fullName()[15:],b.a2.fullName()[15:]
    if hasattr(obj, 'subgroups') and obj.subgroups:
	for grp in obj.subgroups:
	    bondlist2 = propagateBondDataUp(grp)
	    #print 'add',len(bondlist2),'from',hex(id(grp)),'to',hex(id(obj))
	    for b in bondlist2:
		if b.a2 is None:
		    continue	# needed after replaceAtom; must be better way
		#print hex(id(b)),b.a1.fullName()[15:],b.a2.fullName()[15:]
		if not have_bond[id(b)]:
		    have_bond[id(b)] = 1
		    bondlist.append(b)
    try:
	obj.joining_bonds = obj.bonds
    except AttributeError:
	pass
    obj.bonds = bondlist
    return bondlist

def propagateAtomDataUp1(atomlist):
    subgroup_seen = DictWithDefault.DictWithDefault(0)
    for a in atomlist:
	s = a.parent_subgroup
	if not subgroup_seen[id(s)]:
	    subgroup_seen[id(s)] = 1
	    s.atoms = []
	s.atoms.append(a)

def propagateAtomDataUp(obj):
    if hasattr(obj, 'subgroups') and obj.subgroups:
	atoms = []
	for grp in obj.subgroups:
	    atoms = atoms + propagateAtomDataUp(grp)
	if len(obj.atoms) > len(atoms):
	    for a in obj.atoms:
		if not a in atoms:
		    atoms.append(a)
		    print 'kludge propagateAtomDataUp',a
	obj.atoms = atoms
    #print 'propagateAtomDataUp',hex(id(obj)),len(obj.atoms)
    return obj.atoms

def unPropagateDataUp(obj):
    try:
	obj.bonds = obj.joining_bonds
	print 'unPropagateDataUp',hex(id(obj)),obj.bonds
    except AttributeError:
	pass
    if hasattr(obj, 'subgroups'):
	for grp in obj.subgroups:
	    unPropagateDataUp(grp)

def _replace_ba(key, new_atoms):
    k = new_atoms[key]
    if k is not None:
	return id(k)
    return key

def replaceBondAngleLists(atomlist, new_atoms):
    for a in atomlist:
	if hasattr(a, 'bond_angles') and new_atoms[id(a)] is not None:
	    ba_list = DictWithDefault.DictWithDefault(None)
	    for (key, ba) in a.bond_angles.items():
		key2 = (_replace_ba(key[0], new_atoms), _replace_ba(key[1], new_atoms))
		if key2[0] > key2[1]:
		    key2 = (key2[1], key2[0])
		ba_list[key2] = ba
	    new_atoms[id(a)].bond_angles = ba_list
	    print new_atoms[id(a)].fullName(),'replaceBondAngleLists',len(ba_list.keys())

def replaceAtom(blist, new_atoms, replaced_bonds = None, depth = 1):
    if hasattr(blist, 'subgroups'):
	#print ' '*depth, 'replaceAtom',hex(id(blist))
	if hasattr(blist, 'bonds'):
	    replaceAtom(blist.bonds, new_atoms, replaced_bonds, depth+1)
	if hasattr(blist, 'joining_bonds'):
	    replaceAtom(blist.joining_bonds, new_atoms, replaced_bonds, depth+1)
	if hasattr(blist, 'atoms'):
	    for pos in range(len(blist.atoms)):
		new_a = new_atoms[id(blist.atoms[pos])]
		if new_a is not None:
		    blist.atoms[pos] = new_a
	for grp in blist.subgroups:
	    replaceAtom(grp, new_atoms, replaced_bonds, depth+1)
	#print ' '*depth, 'replaceAtom',hex(id(blist))
	return
    for pos in range(len(blist)):
	b = blist[pos]
	#print b
	if not Bonds.isBond(b) and not Bonds.isDihedral(b):
	    replaceAtom(b, new_atoms, replaced_bonds)
	elif new_atoms[id(b.a1)]:
	    new_b = b.replaceAtom(new_atoms)
	    if 0 and Bonds.isDihedral(b):
		print pos,'replaceAtom',hex(id(b)),b,hex(id(new_b)),new_b,hex(id(replaced_bonds))
	    blist[pos] = new_b
	    if replaced_bonds is not None:
		replaced_bonds[id(b)] = new_b
		#print hex(id(new_b)),'replaces',hex(id(b)),b.a1.fullName()[8:]

def _printblist(dlist, depth = 0):
    if Bonds.isBond(dlist[0]):
	for b in dlist:
	    print ' '*depth,'pbond',b.a1.fullName(),b.a2.fullName(),hex(id(b.a1)),hex(id(b.a2)),hex(id(b))
	return
    for d in dlist:
	try:
	    d[0]
	    _printblist(d, depth + 1)
	except: print '_printblist unknown type',type(d),d.__methods__


def replaceBond(dlist, new_bonds, debug = 0):
    if hasattr(dlist, 'subgroups'):
	#print 'replaceBond',hex(id(dlist))
	if hasattr(dlist, 'bonds'):
	    replaceBond(dlist.bonds, new_bonds)
	if hasattr(dlist, 'danglingbonds'):
	    replaceBond(dlist.danglingbonds, new_bonds)
	for grp in dlist.subgroups:
	    replaceBond(grp, new_bonds)
	#if hasattr(dlist, 'original_subgroups'):
	#    for grp in dlist.original_subgroups:
	#	replaceBond(grp, new_bonds)
	return
    for pos in range(len(dlist)):
	d = dlist[pos]
	if Bonds.isDihedral(d):
	    if new_bonds[id(d.b1)] or new_bonds[id(d.b2)] or new_bonds[id(d.b3)]:
		dlist[pos] = d.replaceBond(new_bonds)
		#print 'replace_d',hex(id(d)),'with',hex(id(dlist[pos])),dlist[pos]
	elif Bonds.isBond(d):
	    b = new_bonds[id(d)]
	    if debug: print 'replaceB',hex(id(d)),'with',hex(id(b)),hex(id(new_bonds))
	    if b is not None:
		dlist[pos] = b
	else:
	    replaceBond(d, new_bonds)


def replaceHookBonds(hooks, new_bonds):
    if type(hooks) is type([]):
	for h in hooks:
	    replaceHookBonds(h, new_bonds)
    else:
	replaceBond(hooks, new_bonds)

def _addDihedral(dlist, dbond, new_atom):
    if dbond.constraint is not None:
	d = dbond.constraint
	args = [d.a1, d.a2, d.a3, d.a4, d.angle]
	args[args.index(None)] = new_atom
	dlist.append(Bonds.Dihedral(args))

def _hasSubgoup(new_atom, subgroup1):
    if not hasattr(new_atom, 'parent_subgroup'):
	return 0
    for sg in subgroup1.subgroups:
	if new_atom.parent_subgroup is sg:
	    return 1
	if _hasSubgoup(new_atom, sg):
	    return 1
    return 0


def delSubgroups(group):
    if not hasattr(group,'subgroups'):
	print 'delSubgroups Error!!!! (deleted twice?)',fullSublistName(group)
	print group
	raise RuntimeError
    for sub in group.subgroups:
	delSubgroups(sub)
    delattr(group, 'parent')
    print 'delSubgroups',hex(id(group))
    delattr(group, 'subgroups')

def _closeNestedList(list1):
    try:
	n = len(list1)
    except:
	return list1
    newlist = []
    for i in range(n):
	newlist.append(_closeNestedList(list1[i]))
    return newlist

def _printsublist(obj, depth = 0):
    for sg in obj.subgroups:
	print '  '*depth,fullSublistName(sg)
	_printsublist(sg, depth + 1)

def fullSublistName(obj):
    name = obj.name + ':' + hex(id(obj))
    while obj.parent is not None:
	obj = obj.parent
	try:
	    name1 = obj.name
	except AttributeError:
	    name1 = '?'
	name = name1 + '.' + name + ',' + hex(id(obj))
    return name

def _checkSublistConsistency(obj):
    parent = obj.parent
    while parent is not None:
	print 'check',fullSublistName(parent),'\n',fullSublistName(obj),'\n',parent.subgroups
	_printsublist(parent)
	index = parent.subgroups.index(obj)
	obj = parent
	parent = parent.parent

def _find_equiv_subgroup(new_subgroup, orig_parent_subgroup, newSubgroupMap,\
			 verifyGroup = 1):
    obj = orig_parent_subgroup
    if newSubgroupMap:
	cnt = 0
	while newSubgroupMap.has_key(id(obj)):
	    cnt = cnt + 1
	    if cnt > 100:
		print 'stuck in loop',hex(id(obj)),obj.name
		if cnt > 111:
		    raise RuntimeError
	    obj = newSubgroupMap[id(obj)]
	if obj is not orig_parent_subgroup:
	    if not verifyGroup:
		return obj
	    parent = obj
	    try:
		while parent is not None:
		    #print 'parent.name',hex(id(parent)),parent.name
		    if parent is new_subgroup:
		    #print 'mapped',hex(id(orig_parent_subgroup)),orig_parent_subgroup.name,'to',hex(id(obj)),obj.name
			return obj
		    parent = parent.parent
	    except AttributeError:
		pass
	    o = orig_parent_subgroup
	    while newSubgroupMap.has_key((id(o), id(o))):
		o = newSubgroupMap[(id(o), id(o))]
	    return o
	    #print 'map not applied'
	    obj = orig_parent_subgroup
	else:
	    #print 'warning-no mapping for',hex(id(orig_parent_subgroup)),hex(id(newSubgroupMap)),hex(id(None))
	    #print orig_parent_subgroup
	    return orig_parent_subgroup
		# not in newSubgroupMap, figure out by comparing place in tree
    parent = orig_parent_subgroup.parent
    path = []
    while parent is not None:
	#print 'ascend',parent.name,len(parent.subgroups),hex(id(obj)),fullSublistName(obj)
	index = parent.subgroups.index(obj)
	#print 'ascend',index,parent.name,len(parent.subgroups)
	path.append(index)
	obj = parent
	parent = parent.parent
    path.reverse()
    #pprint.pprint(path)
    equiv_subgroup = new_subgroup
    for index in path:
	#print 'descend',index,equiv_subgroup.name,len(equiv_subgroup.subgroups)
	equiv_subgroup = equiv_subgroup.subgroups[index]
    if 0:
	o = orig_parent_subgroup
	while newSubgroupMap.has_key((id(o), id(o))):
	    o = newSubgroupMap[(id(o), id(o))]
	if equiv_subgroup is not o:
	    print 'equiv_subgroup',hex(id(equiv_subgroup)),'vs',hex(id(o))
    return equiv_subgroup

def filterHooks(hooks, newGroup, topGroup, hookUsed, new_hook, \
		newSubgroupMap, do_clone = 1):
    do_clone2 = do_clone
    #print 'enter filterHooks',do_clone2,hooks
    if do_clone2:
	newlist = HookList([])
    else:
	newlist = hooks
    for i in range(len(hooks)):
	h = hooks[i]
	#print i,':',h
	if type(h) is type([]) or isHookList(h):
	    try:
		a = h[0].getBonds()[0].a1
	    except AttributeError, msg:
		print 'AttributeError',msg
		print 'in filterHooks',hex(id(h[0])),h[0].getBonds()
		continue
	    #print '_filterHooks recurse',hex(id(a)),a.name
	    subgroup = _find_equiv_subgroup(topGroup, a.parent_subgroup, \
					    newSubgroupMap, do_clone)
	    r = filterHooks(h, subgroup, topGroup, hookUsed, new_hook, \
			    newSubgroupMap, do_clone)
	    if do_clone2:
		newlist.append(r)
	    else:
		newlist[i] = r
	else:
	    if h is hookUsed:
		continue
	    if do_clone:
		dblist = h.getBonds()[:]
		#print 'filterHooks',hex(id(h)),dblist
		newHook = Hook((newGroup, dblist))
		newlist.append(newHook)
	    else:
		h._group = newGroup
		newHook = h
		if do_clone2:
		    newlist.append(newHook)
	    new_hook[id(h)] = newHook

	    #	    if type(h) is type([]):
	    #		for h2 in h:
	    #		    new_hook[id(h2)] = newHook
	    #	    else:
    if not do_clone2:
	try:
	    newlist.remove(hookUsed)
	except ValueError: pass# print 'warn',hex(id(hookUsed)),'not in hooks',hookUsed
    #print 'filterHooks done',len(newlist)
    return newlist

def _containsGroup(search_in_group, search_for_group):
    for grp in search_in_group.subgroups:
	if search_for_group is grp:
	    return 1
	if _containsGroup(grp, search_for_group):
	    return 1
    return 0

def _printAttachments(grp, depth = 1):
    try:
	for a in grp.attach_subgroups:
	    print '   '*depth, hex(id(grp)),'attach',fullSublistName(a)
    except AttributeError:
	print '   '*depth, hex(id(grp))
    for subgrp in grp.subgroups:
	_printAttachments(subgrp, depth+1)


def _lowerAttachedGroups(subgrp, removed_subgrp, collapsed_map):
    collapsed_to = collapsed_map[id(subgrp)]
    if collapsed_to:
	for s in subgrp.subgroups:
	    _lowerAttachedGroups(s, removed_subgrp, collapsed_map)
	return
    if 0:
	print '_reDoAttachedGroups',len(removed_subgrp.attach_subgroups)
	print fullSublistName(subgrp)
	print fullSublistName(removed_subgrp)
	for t in removed_subgrp.attach_subgroups:
	    print '  ',fullSublistName(t)
    #except AttributeError: pass
    try:
	list1 = subgrp.attach_subgroups
    except AttributeError:
	list1 = []
    try:
	list2 = removed_subgrp.attach_subgroups
    except AttributeError:
	list2 = []
    subgrp.attach_subgroups = list1 + list2

def _reDoAttachedGroups(parent, removed_subgrp, collapsed_map):
    past_removed = 0
    for s in parent.subgroups:
	if s is removed_subgrp:
	    for s2 in removed_subgrp.subgroups:
		_lowerAttachedGroups(s2, removed_subgrp, collapsed_map)
	    past_removed = 1
	elif past_removed:
	    try:
		pos = s.attach_subgroups.index(removed_subgrp)
	    except ValueError:
		pass #print '_reDoAttachedGroups warning, no',hex(id(removed_subgrp)),'in',hex(id(s))
	    else:
		s.attach_subgroups.remove(removed_subgrp)
		_replaceAttachedGroups(s, removed_subgrp, collapsed_map)
	else: pass #print '_reDoAttachedGroups warning,',hex(id(s)),'precedes'

def _replaceAttachedGroups(subgrp, removed_subgrp, collapsed_map):
    #print 'replaceAttachedGrp',hex(id(removed_subgrp)),'from',hex(id(subgrp))
    for sub in removed_subgrp.subgroups:
	if collapsed_map[id(sub)]:
	    _replaceAttachedGroups(subgrp, sub, collapsed_map)
	else:
	    subgrp.attach_subgroups.append(sub)
