
import hull
import DelaunayTriangulation
import umath
from Vector import Vector
from Geometry import Line, LineSegment, Plane, Sphere, Polygon, Prism, Circle
from ChemicalObjects import Group
from CompositeObject import translateAtomsBy
from DictWithDefault import DictWithDefault
import Units, Utility
import pprint
from time import clock

def PositionAdjacent(obj1, obj1Atoms, obj2, obj2Atoms, distances):
    t0 = clock()
    n = len(obj1Atoms)
    assert(n == len(obj2Atoms))
    assert(n <= len(distances))
    if n == 3:
	return PositionAdjacent3(obj1, obj1Atoms, obj2, obj2Atoms, distances)
    desired_pos2 = _FirstPairPosition(obj1, obj1Atoms[0], obj2, distances[0])
    if not desired_pos2:
	return
    t1 = clock()
    obj2.translateBy(desired_pos2 - obj2Atoms[0].position())
    desired_pos1 = _FirstPairPosition(obj2, obj2Atoms[0], obj1, distances[0])
    v2 = desired_pos1 - desired_pos2
    v1 = obj1Atoms[0].position() - obj2Atoms[0].position()
    rotate_angle = v2.angle(v1)
    axis = Line(obj2Atoms[0].position(), v2.cross(v1))
    print 'rotate_angle',rotate_angle/Units.deg, axis, v1, v2,\
	  '%.3f' % (obj2Atoms[0].position() - obj1Atoms[0].position()).length()
    obj2.RotateAtomsOnOneSide(obj2Atoms[0], None, axis, rotate_angle)
    print 'PositionAdjacent times',t1 - t0,clock() - t1
    print 'new dist %.3f' % (obj2Atoms[0].position() - obj1Atoms[0].position()).length()

def printAtomsPairsByDistance(atom_list1, atom_list2):
    dmap = {}
    atom_list = atom_list1 + atom_list2
    listAtomsNear(atom_list, atom_list2[0], None, dmap)
    closest_pair_dist = 1.e9
    for a in atom_list2:
	for a2 in dmap[id(a)]:
	    if a2 == -1:
		continue
	    d = (a.position() - a2.position()).length()
	    if d < closest_pair_dist:
		closest_pair_dist = d
		closest_pair = (a2, a)
    print 'closest pair %.3f' % closest_pair_dist, closest_pair[0].fullName(), closest_pair[1].fullName()
    out_lines = []
    for a in atom_list:
	for a2 in dmap[id(a)]:
	    try:
		l = 'distance %.3f to %s %s' %\
		    ((a.position() - a2.position()).length(), \
		     a.fullName(), a2.fullName())
		out_lines.append(l)
	    except AttributeError:
		print 'err in printAtomsPairsByDistance',a,a2
    out_lines.sort()
    for l in out_lines[:999]: print l

def listAtomsNear(atoms_in_area, atom1, delaunay_list = None, delaunay_map = None):
    pos_list = []
    a_list = []
    a_index = -1
    for a in atoms_in_area:
	if a is atom1:
	    a_index = len(pos_list)
	pos = a.position()
	if pos is not None:
	    a_list.append(a)
	    pos_list.append(pos)
    print 'listAtomsNear',atoms_in_area
    print a_index, atom1
    assert(a_index >= 0)
    triang_index = hull.Delaunay(pos_list, 1000)
    atoms_near = []
    create_map = (delaunay_map is not None and not delaunay_map)
    create_list = (delaunay_list is not None and not delaunay_list)
    for t in triang_index:
	if a_index in t:
	    for index in t:
		if index != -1 and index != a_index:
		    atoms_near.append(a_list[index])
	if create_list:
	    delaunay_list.append(map(lambda index,l = a_list :\
				     (l[index], -1)[index == -1], t))
	    #print 'delaunay_list',delaunay_list[-1],t
	if create_map:
	    for index in t:
		if index == -1:
		    continue
		try:
		    l = delaunay_map[id(a_list[index])]
		except KeyError:
		    l = []
		for index2 in t:
		    if index2 != index:
			if index2 == -1:
			    l.append(index2)
			else:
			    l.append(a_list[index2])
		delaunay_map[id(a_list[index])] = l
    return atoms_near

def _FirstPairPosition(obj1, obj1Atom, obj2, distance):
    (bbox_min, bbox_max) = obj2.boundingBox()
    obj_max_dim = (bbox_min - bbox_max).length()
    max_dim = distance + 0.3 + obj_max_dim
    neighbors = {}
    delaunay_list = []
    atoms_in_area = obj1.listAtomsWithinDistance(obj1Atom.position(), max_dim)
    atoms_near = listAtomsNear(atoms_in_area, obj1Atom, delaunay_list, neighbors)
    if not atoms_near:
	return Vector(distance, 0, 0)	# no constraint - choose arbitrary

    region_outside_hull = DictWithDefault(0)
    hull_tetrahedra = []
    vertex_on_hull = DictWithDefault(0)
    tetrahedra_containing = DictWithDefault([])
    tetrahedra_containing3 = DictWithDefault([])
    most_edges = 0
    atom_with_most_edges = None
    for tetrahedron in delaunay_list:
	if -1 in tetrahedron:
	    region_outside_hull[id(tetrahedron)] = tetrahedron
	    hull_tetrahedra.append((tetrahedron, -1))
	    for a in tetrahedron:
		if a != -1:
		    vertex_on_hull[id(a)] = 1
	for a in tetrahedron:
	    list3 = tetrahedron[:]
	    list3.remove(a)
	    if -1 not in list3:
		list3.sort(_cmpAtoms)
		tetrahedra_containing3[tuple(list3)].append(tetrahedron)
	    if a != -1:
		tetrahedra_containing[id(a)].append(tetrahedron)
		if len(tetrahedra_containing[id(a)]) > most_edges:
		    most_edges = len(tetrahedra_containing[id(a)])
		    atom_with_most_edges = a
    if 0 and atom_with_most_edges:
	_display_edges_to(atom_with_most_edges, tetrahedra_containing[id(a)], obj1)
    while hull_tetrahedra:
	best_eval = 0
	best_region = None
	for (tetrahedron, exclude) in hull_tetrahedra:
	    (eval, x) = _eval_tetrahedron(tetrahedron, obj1Atom, \
					  distance, exclude, best_eval, obj2)
	    if eval > best_eval or x is tetrahedron:
		print 'eval',obj1Atom,eval,x,'best_region',best_region
		best_eval = eval
		if x is tetrahedron:
		    if type(best_region) is type([]):
			best_region.append(tetrahedron)
		    else: best_region = [tetrahedron]
		else:
		    best_region = (tetrahedron, x)
	if type(best_region) is type([]):
	    return _merge_regions(best_region, obj1Atom, distance)
	next_layer = []
	for (tetrahedron, odd_vertex) in hull_tetrahedra:
	    _addAdjacentTetrahedra(next_layer, tetrahedron, odd_vertex, \
				   tetrahedra_containing3)
	hull_tetrahedra = next_layer
	print 'next_layer',best_region
	pprint.pprint(next_layer)
    if not best_region:
	return None
    return best_region[1]

def _cmpAtoms(a1, a2):
    return cmp(id(a1), id(a2))

def _merge_regions(tetrahedra, obj1Atom, distance):
    pos_list = []
    atom_used = DictWithDefault(0)
    atom_used[id(obj1Atom)] = 1
    atom_used[id(-1)] = 1
    for tetrahedron in tetrahedra:
	for a in tetrahedron:
	    if not atom_used[id(a)]:
		pos_list.append(a.position())
		atom_used[id(a)] = 1
	    if a != -1:
		print '_merge_regions',len(pos_list),a.name,tetrahedron
    assert(len(pos_list) >= 3)
    o1pos = obj1Atom.position()
    best_sum = 0
    best3pos = None
    for pos1 in pos_list:		# O(N^3) algorithm, should be O(NlogN)?
	v1 = pos1 - o1pos
	for pos2 in pos_list:
	    v2 = pos2 - o1pos
	    angle12 = v1.angle(v2)
	    for pos3 in pos_list:
		v3 = pos3 - o1pos
		sum = angle12 + v3.angle(v2) + v3.angle(v1)
		if sum > best_sum:
		    best_sum = sum
		    best3pos = (pos1, pos2, pos3)
    plane = Plane(best3pos[0], best3pos[1], best3pos[2])
    dir = plane.normal
    if dir * (o1pos - best3pos[0]) < 0:
	dir = -dir
    return o1pos + distance*dir

def _eval_tetrahedron(tetrahedron, obj1Atom, distance, odd_vertex, best_eval, obj2):
    if odd_vertex == -1 and (obj1Atom in tetrahedron):
	return (1.e9, tetrahedron)
    trisect_line = line_trisecting_facet_excluding(tetrahedron, odd_vertex)
    if not trisect_line:
	return (0, None)
    pos1 = obj1Atom.position()
    if (pos1 - trisect_line.point).length() > distance:
	return (0, None)
    x = trisect_line.intersectWith(Sphere(pos1, distance))
    if not x:
	return (0, None)
    if (x[0] - pos1) * (trisect_line.point - pos1) > 0:
	intersect_pt = x[0]
    else:
	intersect_pt = x[1]
    for a in tetrahedron:
	if a != -1 and a is not obj1Atom:
	    pt = a.position()
	    break
    length = (trisect_line.point - pt).length()
    if length < 0.15:	# no group could fit thru such a small triangle
	print 'tetrahedron',tetrahedron,'too small',length
	return (0, None)
    if 0 and length > best_eval:
	length2 = _eval_fit(tetrahedron, trisect_line, obj1Atom, obj2, distance)
    print '_eval_tetrahedron',length,best_eval,'goal %.3f' % distance,\
	  'how far from goal? %.3f' % (intersect_pt - obj1Atom.position()).length()
    return (length, intersect_pt)

def _eval_fit(tetrahedron, trisect_line, obj1Atom, obj2, obj2Atom,\
	      candidate_pos, distance):
    atom_pos = atom.position()
    pos_list = []
    a_list = []
    for a in obj.atoms:
	pos = a.position()
	if Utility.isDefinedPosition(pos):
	    pos_list.append(pos)
	    a_list.append(a)
    chull = hull.ConvexHull(pos_list, 1000)

#    triangle = tetrahedron[:]
#    triangle.remove(?)
#    prism_base = Polygon(triangle)
#    prism = Prism(prism_base, trisect_line.point - atom_pos)
#    ht = (candidate_pos - trisect_line.point).length()
#    for axis in search_axes:
#	v0 = axis.cross(randomDirection())
#	for u in arange(0, 2*umath.pi, umath.pi/64):
#	    v = rotateDirection(v0, axis, u)
#	    base = prism_base.moveToFace(Line(obj2Atom.position()-ht*v, v))
#	    prism = Prism(base, ht + ?)
	    
    #cross_section = crossSectionAt(obj2, distance, obj1Atom)



    
def line_trisecting_facet_excluding(tetrahedron, exclude_obj):
    triangle_atoms = []
    for o in tetrahedron:
	if o != exclude_obj:
	    triangle_atoms.append(o)
    #print 'line_trisecting_facet_excluding',tetrahedron,triangle_atoms,exclude_obj
    assert(len(triangle_atoms) == 3)
    line1 = LineSegment(triangle_atoms[0].position(), triangle_atoms[1].position())
    line2 = LineSegment(triangle_atoms[1].position(), triangle_atoms[2].position())
    tri_plane = Plane(line1.point, line1.point2, line2.point2)
    bisect1 = line1.PerpBisector(tri_plane)
    bisect2 = line2.PerpBisector(tri_plane)
    pt = bisect1.intersectWith(bisect2)
    #print line1
    #print line2
    #print 'angle',line1.direction.angle(line2.direction)/Units.deg
    #print 'bisects',bisect1,bisect2,pt
    if not pt:				# degenerate triangle?
	return None
    return Line(pt, bisect1.direction.cross(bisect2.direction))

def _display_edges_to(atom, tetrahedra_list, group):
    import RasMol
    ras_session = RasMol.RasMol(group, 0)
    for tet in tetrahedra_list:
	coords = []
	for a in tet:
	    if a != -1:
		coords.append(a.position())
		if a is not atom:
		    print '%.2f [%5.2f,%5.2f,%5.2f] %s' % \
			  ((coords[-1] - atom.position()).length(),
			   coords[-1][0],coords[-1][1],coords[-1][2],
			   a.fullName())
	ras_session.AddPolygon(coords)
    ras_session.RefreshScreen()
    ras_session.Interact()
    raise RuntimeError

    

def _addAdjacentTetrahedra(next_layer, tetrahedron, odd_vertex, neighbors3):
    min_length = 0.2
    if odd_vertex == -1:
	list = tetrahedron[:]
	list.remove(odd_vertex)
	if (list[0].position() - list[1].position()).length() < min_length and\
	   (list[0].position() - list[2].position()).length() < min_length and\
	   (list[1].position() - list[2].position()).length() < min_length:
	    return
	list.sort(_cmpAtoms)
	list3 = neighbors3[tuple(list)][:]
	list3.remove(tetrahedron)
	#print 'tetrahedra_containing',list[0],list[1],list[2]
	if len(list3) == 1:
	    for a in list3[0]:
		if a not in list3:
		    odd = a
	    if odd != -1:
		next_layer.append((list3[0], odd))
	    return
	raise RuntimeError,'internal error; no tetrahedron found'
    

#
# check to see if other atoms can fit through _addAdjacentTetrahedra?
#
    


def findNearestPointOnSurface(obj, atom):
    dt = DelaunayTriangulation.ConvexHull(obj)
    return dt.nearestPointOnSurface(atom.position())

def positionOnPlane(plane_points, obj, atoms):
     translate_v = plane_points[0] - atoms[0].position()
     obj.translateBy(translate_v)
     v2 = atoms[1].position() - atoms[0].position()
     pt2 = plane_points[1]
     desired_v2 = pt2 - atoms[0].position()
     axis = desired_v2.cross(v2)
     angle = v2.angle(desired_v2)
     obj.rotateAroundPoint(axis, angle, plane_points[0])
     print 'got',desired_v2.angle(atoms[1].position() - atoms[0].position())/Units.deg
     print 'positionOnPlane d2',(plane_points[0] - atoms[0].position()).length(),(plane_points[1] - atoms[1].position()).length()
     print 'p1-plane',Plane(plane_points[0],plane_points[1],plane_points[2]).distanceFrom(atoms[1].position()),desired_v2.angle(atoms[1].position() - atoms[0].position())/Units.deg,angle/Units.deg
     v3 = atoms[2].position() - plane_points[0]
     pt3 = plane_points[2]
     axis = plane_points[0] - plane_points[1]
     axis_line = Line(plane_points[1], axis)
     pivot = axis_line.projectionOf(pt3)
     desired_v3 = pt3 - pivot
     projection = axis_line.projectionOf(atoms[2].position())
     v3 = atoms[2].position() - projection
     angle = v3.angle(desired_v3)
     obj.rotateAroundPoint(axis, angle, plane_points[0])
     print 'got angle', desired_v3.angle(atoms[2].position() - projection)/Units.deg
     print 'positionOnPlane d3',(plane_points[0] - atoms[0].position()).length(),(plane_points[1] - atoms[1].position()).length(),(plane_points[2] - atoms[1].position()).length()

def optimizePositionInPlane(obj2, plane_pts, obj2Atoms, obj1Atoms, distances):
    want_triangle = Polygon(plane_pts)
    current_triangle = Polygon(map(lambda a : a.position(), obj2Atoms))
    delta = want_triangle.center() - current_triangle.center()
    obj2.translateBy(delta)
    sum_dist = 0
    for i in range(3):
	o1pos = obj1Atoms[i].position()
	o2pos = obj2Atoms[i].position()
	normal = (plane_pts[i] - o1pos).cross(o2pos - o1pos)
	circle = Circle(o1pos, normal, distances[i])
	plane2 = Plane(o2pos, plane_pts[i] - o2pos)
	pts = circle.intersectWith(plane2)
	if pts is None:
	    d = (plane2.distanceFrom(circle.center) - distances[i])/Units.Ang
	    print 'Error - atoms on 2nd object not in similar enough shape',\
		  ' to 1st object atoms\nat atom pair number',i,\
		  'discrepancy %.3f Angstroms' % d
	    return
	dist1 = (pts[0] - o2pos).length()
	dist2 = (pts[1] - o2pos).length()
	if dist1 < dist2:
	    sum_dist = sum_dist + dist1
	else:
	    sum_dist = sum_dist + dist2
    if current_triangle.normal * (o2pos - o1pos) > 0:
	sign = -1
    else:
	sign = 1
    delta2 = (sign * sum_dist / 3) * current_triangle.normal
    obj2.translateBy(delta2)
    print 'center by',delta.length(),'lower',delta2.length()
    

def PositionAdjacent3(obj1, obj1Atoms, obj2, obj2Atoms, distances):
    if len(distances) == 1:
	distances = distances * len(obj1Atoms)
    else:
	assert(len(distances) == len(obj1Atoms))
    spheres1 = []
    for i in range(3):
	a = obj1Atoms[i]
	spheres1.append(Sphere(a.position(), distances[i]))
    tangent = [[], [], []]   # circles of lines tangent to each pair of spheres
    for i in range(0, 2):
	for j in range(i + 1, 3):
	    normal = (spheres1[j].center - spheres1[i].center).normal()
	    sin_angle = (spheres1[j].radius - spheres1[i].radius) / \
			   (spheres1[j].center - spheres1[i].center).length()
	    centeri = spheres1[i].center - (sin_angle*spheres1[i].radius)*normal
	    centerj = spheres1[j].center - (sin_angle*spheres1[j].radius)*normal
	    angle1 = umath.arcsin(sin_angle)
	    cos_angle = umath.cos(angle1)
	    tangent[i].append(Circle(centeri, normal, cos_angle*spheres1[i].radius))
	    tangent[j].append(Circle(centerj, normal, cos_angle*spheres1[j].radius))
    plane_centers = Plane(obj1Atoms[0].position(), obj1Atoms[1].position(), obj1Atoms[2].position())
    plane1_points = []
    plane2_points = []
    for i in range(3):
	pt_pair = tangent[i][0].intersectWith(tangent[i][1])
	assert(len(pt_pair) == 2)
	if plane_centers.normal * (pt_pair[0] - spheres1[i].center) > 0:
	    plane1_points.append(pt_pair[0])
	    plane2_points.append(pt_pair[1])
	else:
	    plane1_points.append(pt_pair[1])
	    plane2_points.append(pt_pair[0])
    delaunay = DelaunayTriangulation.DelaunayTriangulation(obj1)
    delaunay.removeFacesBiggerThan(4*Units.Ang*Units.Ang)
    eval1 = 0
    eval2 = 0
    conv_hull = DelaunayTriangulation.ConvexHull(obj1)
    for i in range(3):
	tet1index = delaunay.contains(plane1_points[i])
	eval1 = eval1 + (not not tet1index)
	tet2index = delaunay.contains(plane2_points[i])
	eval2 = eval2 + (not not tet2index)

	path1 = []
	near1 = conv_hull.nearestPointOnSurface(plane1_points[i], path1)
	path2 = []
	near2 = conv_hull.nearestPointOnSurface(plane2_points[i], path2)
	print 'eval1 %.3f %.3f' % (eval1,(near1 - plane1_points[i]).length()),\
	      'eval2 %.3f %.3f' % (eval2,(near2 - plane2_points[i]).length())
	for t in path1:
	    print 'path1 area',t.area()
	for t in path2:
	    print 'path2 area',t.area()
	
    if 1: # eval1 == 0 and eval2 == 0:
	#for i in range(3):
	#    pts1.append(findNearestPointOnSurface(obj1, obj1Atoms[i]))
	#    pts2.append(findNearestPointOnSurface(obj2, obj2Atoms[i]))
	    
	cm = obj1.centerOfMass()
	if plane_centers.normal * (cm - spheres1[0].center) < 0:
	    eval1 = 1
	else:
	    eval2 = 1
	print 'use cm',eval1, eval2, plane_centers.normal,cm,spheres1[0].center
    if eval1 < eval2:
	plane_points = plane1_points
    else:
	plane_points = plane2_points
    coord_list = [plane1_points + [plane1_points[0]] + plane2_points]
    positionOnPlane(plane_points, obj2, obj2Atoms)
    for i in range(3):
	o2pos = obj2Atoms[i].position()
	print i,'pos_dist %.3f %.3f %.3f' % ((obj2Atoms[i].position() - obj1Atoms[i].position()).length(),\
	      (obj2Atoms[i].position() - plane_points[i]).length(),\
	      (obj2Atoms[i].position() - obj2Atoms[(i+1)%3].position()).length()),\
	      '[%.2f,%.2f,%.2f]' % (o2pos[0], o2pos[1], o2pos[2]),\
	      obj1Atoms[i].fullName()[-8:],obj2Atoms[i].fullName()[-8:]
    if 1:
	try:
	    p2 = Plane(plane2_points[0], plane2_points[1], plane2_points[2])
	    o2 = Plane(obj2Atoms[0].position(), obj2Atoms[1].position(), obj2Atoms[2].position())
	    print 'plane angles %.2f %.3f' % (plane_centers.normal.angle(p2.normal)/Units.deg,\
					      o2.normal.angle(p2.normal)/Units.deg)
	except ZeroDivisionError:
	    print 'ERROR points dont form plane'
	    return coord_list
    coord_list.append(map(lambda a: a.position(), obj2Atoms + [obj2Atoms[0]]))
    optimizePositionInPlane(obj2, plane_points, obj2Atoms, obj1Atoms, distances)
    for i in range(3):
	print i,'pos_dist %.2f %.2f   %.2f ' %\
	      ((obj2Atoms[i].position() - obj1Atoms[i].position()).length(),\
	       (obj2Atoms[i].position() - plane_points[i]).length(),\
	       (plane_points[i] - plane_points[(i+1) % 3]).length()),\
	      obj1Atoms[i].fullName()[-8:],obj2Atoms[i].fullName()[-8:]
    if 1:
	p2 = Plane(plane2_points[0], plane2_points[1], plane2_points[2])
	o2 = Plane(obj2Atoms[0].position(), obj2Atoms[1].position(), obj2Atoms[2].position())
	print 'plane angles %.2f %.3f' % (plane_centers.normal.angle(p2.normal)/Units.deg,\
					  o2.normal.angle(p2.normal)/Units.deg)
    coord_list.append(map(lambda a: a.position(), obj2Atoms + [obj2Atoms[0]]))
    return coord_list
