#ifndef PIXUTILS_H_INCLUDED
#define PIXUTILS_H_INCLUDED
/* pixutils.h
 * RasMol2 Molecular Graphics
 * Roger Sayle, August 1995
 * Version 2.6
 */

typedef struct {
        int px, py, pz;      /* Spline Control Co-ordinate */
        int tx, ty, tz;      /* Spline Direction Vector    */
        int hnx, hny, hnz;   /* Horizontal Normal Vector   */
        int vnx, vny, vnz;   /* Vertical Normal Vector     */
        int dx, dy, dz;      /* Ribbon Height Vector       */
        int wx, wy, wz;      /* Ribbon Width Vector        */
        char hinten;         /* Horizontal Intensity       */
        char vinten;         /* Vertical Intensity         */
        short hsize;         /* Horizontal Vector Length   */
        short vsize;         /* Vertical Vector Length     */
        short wide;          /* Ribbon Width               */
    } Knot;

typedef struct {
        Pixel __huge *fbuf;
        short __huge *dbuf;
        int xmax, ymax;
        int yskip;
    } ViewStruct;

#define ZValid(z)     ((!UseSlabPlane) || ((z)<SlabValue))
#define XValid(x, xmax)     (((x)>=0)&&((x)<(xmax)))
#define YValid(y, ymax)     (((y)>=0)&&((y)<(ymax)))

#define MAXVERT 256
typedef struct {
                int x, y, z;
                int xorg, yorg, zorg;
                int inten;
               } Vert;

typedef struct {
                Vert v[MAXVERT];
                int count;
               } Poly;


#ifdef PIXUTILS
/* ViewStruct View; */
int SplineCount;
int FontSize;

#else
/* extern ViewStruct View; */
extern int SplineCount;
extern int FontSize;
#endif

#ifdef FUNCPROTO
void PlotDeepPoint( int, int, int, int, ViewStruct *bitmap );
void ClipDeepPoint( int, int, int, int, ViewStruct *bitmap );
void DrawTwinLine( int, int, int, int, int, int, int, int, ViewStruct *bitmap);
void ClipTwinLine( int, int, int, int, int, int, int, int, ViewStruct *bitmap);
void DrawTwinVector( int, int, int, int, int, int, int, int, ViewStruct *bitmap );
void ClipTwinVector( int, int, int, int, int, int, int, int, ViewStruct *bitmap );
void ClipDashVector( int, int, int, int, int, int, int, int, ViewStruct *bitmap);
void DisplayPoly( ViewStruct *bitmap);
Vert *GetVert(int);
int NextVertIndex(int);

void DrawCylinder( int, int, int, int, int, int, int, int, int, ViewStruct *bitmap );
void ClipCylinder( int, int, int, int, int, int, int, int, int, ViewStruct *bitmap );
void DashRibbon( Knot __far*, Knot __far*, int, int, ViewStruct *bitmap );
void StrandRibbon( Knot __far*, Knot __far*, int, int, ViewStruct *bitmap );
void SolidRibbon( Knot __far*, Knot __far*, int, ViewStruct *bitmap );
void RectRibbon( Knot __far*, Knot __far*, int, ViewStruct *bitmap );
void DrawSphere( int, int, int, int, int, ViewStruct *bitmap );
void ClipSphere( int, int, int, int, int, ViewStruct *bitmap );

void SetFontSize( int );
void DisplayString( int, int, int, char*, int, ViewStruct *bitmap );
void InitialisePixUtils(void);

#else /* non-ANSI C compiler */
void PlotDeepPoint();
void ClipDeepPoint();
void DrawTwinLine();
void ClipTwinLine();
void DrawTwinVector();
void ClipTwinVector();
void ClipDashVector();
void DisplayPoly();
Vert *GetVert();
int NextVertIndex();

void DrawCylinder();
void ClipCylinder();
void DashRibbon();
void StrandRibbon();
void SolidRibbon();
void RectRibbon();
void DrawSphere();
void ClipSphere();

void SetFontSize();
void DisplayString();
void InitialisePixUtils();

#endif

#endif /* PIXUTILS_H_INCLUDED */
